"""
nacos for python

"""
from . import func_fuse
from .nacos import Nacos, NacosBalanceClient

__version__ = '1.1.0'

name = 'nacos-starter'

__author__ = 'chizongyang'

__all__ = ['Nacos', 'NacosBalanceClient', 'func_fuse']
