"""
Constants config file.
"""
# pylint: disable=line-too-long
LOG_FORMAT = (
    "%(asctime)s -%(name)s- %(threadName)s-%(thread)d - %(levelname)s - %(message)s"
)
DATE_FORMAT = "%m/%d/%Y %H:%M:%S %p"
TIME_OUT = 10
BEAT_TIME = 10
REGISTER_DICT_KEY = "healthy"
